# =============================================================================
# Ural Facebook-related heuristic functions
# =============================================================================
#
# Collection of functions crafted to work with Facebook's urls.
#
import re

from ural.ensure_protocol import ensure_protocol
from ural.patterns import DOMAIN_TEMPLATE, QUERY_VALUE_IN_URL_TEMPLATE

from ural.utils import (
    safe_parse_qs,
    unquote,
    urljoin,
    pathsplit,
    urlsplit,
    urlunsplit,
    safe_urlsplit,
    SplitResult,
)

NUMERIC_ID_RE = re.compile(r"[0-9]{8,}")

BASE_FACEBOOK_URL = "https://www.facebook.com"

FACEBOOK_ID_RE = re.compile(r"^\d+$")
FACEBOOK_FULL_ID_RE = re.compile(r"^\d+_\d+$")
FACEBOOK_DOMAIN_RE = re.compile(r"(?:facebook\.[^.]+$|fb\.me$)", re.I)
FACEBOOK_URL_RE = re.compile(
    DOMAIN_TEMPLATE % r"(?:[^.]+\.)*(?:facebook\.[^.]+|fb\.me)", re.I
)
MOBILE_REPLACE_RE = re.compile(r"^([^.]+\.)?facebook\.", re.I)

URL_EXTRACT_RE = re.compile(QUERY_VALUE_IN_URL_TEMPLATE % r"u")


def is_facebook_id(value):
    return bool(re.search(FACEBOOK_ID_RE, value))


def is_facebook_full_id(value):
    return bool(re.search(FACEBOOK_FULL_ID_RE, value))


def is_facebook_url(url):
    """
    Function returning whether the given url is a valid Facebook url.

    Args:
        url (str): Url to test.

    Returns:
        bool: Whether given url is from Facebook.

    """
    if isinstance(url, SplitResult):
        return bool(re.search(FACEBOOK_DOMAIN_RE, url.hostname))

    return bool(re.match(FACEBOOK_URL_RE, url))


def is_facebook_post_url(url):
    if not is_facebook_url(url):
        return False

    return (
        "/posts/" in url
        or "/permalink/" in url
        or ("/permalink.php" in url and ("&id=" in url or "&amp;id=" in url))
        or ("/story.php" in url and ("&id=" in url or "&amp;id=" in url))
    )


def is_facebook_link(url):
    splitted = safe_urlsplit(url)

    if not splitted.hostname or ".facebook." not in splitted.hostname:
        return False

    if splitted.path != "/l.php":
        return False

    return True


def extract_url_from_facebook_link(url):
    m = URL_EXTRACT_RE.search(url)

    if m is None:
        return None

    return unquote(m.group(1))


def convert_facebook_url_to_mobile(url):
    """
    Function parsing the given facebook url and returning the same but for
    the mobile website.
    """
    safe_url = ensure_protocol(url)

    has_protocol = safe_url == url

    scheme, netloc, path, query, fragment = urlsplit(safe_url)

    if "facebook" not in netloc:
        raise TypeError(
            "ural.facebook.convert_facebook_url_to_mobile: %s is not a facebook url"
            % url
        )

    netloc = re.sub(MOBILE_REPLACE_RE, "m.facebook.", netloc)

    result = (scheme, netloc, path, query, fragment)

    result = urlunsplit(result)

    if not has_protocol:
        result = result.split("://", 1)[-1]

    return result


class FacebookParsedItem(object):
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False

        for attr in self.__slots__:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        class_name = self.__class__.__name__

        representation = "<" + class_name

        for key in self.__slots__:
            value = getattr(self, key)

            if value is None:
                continue

            representation += " %s=%s" % (key, value)

        representation += ">"

        return representation


class FacebookUser(FacebookParsedItem):
    __slots__ = ("id", "handle")

    def __init__(self, id, handle=None):
        self.id = id
        self.handle = handle

    @property
    def url(self):
        if self.handle is None:
            return urljoin(BASE_FACEBOOK_URL, "/profile.php?id=%s" % self.id)

        return urljoin(BASE_FACEBOOK_URL, "/%s" % self.handle)


class FacebookHandle(FacebookParsedItem):
    __slots__ = ("handle",)

    def __init__(self, handle):
        self.handle = handle

    @property
    def url(self):
        return urljoin(BASE_FACEBOOK_URL, "/%s" % self.handle)


class FacebookGroup(FacebookParsedItem):
    __slots__ = ("id", "handle")

    def __init__(self, id=None, handle=None):
        self.id = id
        self.handle = handle

    @property
    def url(self):
        if self.handle is not None:
            return urljoin(BASE_FACEBOOK_URL, "groups/%s" % self.handle)

        return urljoin(BASE_FACEBOOK_URL, "groups/%s" % self.id)


class FacebookPost(FacebookParsedItem):
    __slots__ = ("id", "parent_id", "parent_handle", "group_id", "group_handle")

    def __init__(
        self,
        post_id,
        parent_id=None,
        parent_handle=None,
        group_id=None,
        group_handle=None,
    ):
        self.id = post_id
        self.parent_id = parent_id
        self.parent_handle = parent_handle
        self.group_id = group_id
        self.group_handle = group_handle

    @property
    def url(self):
        if self.parent_handle is not None:
            return urljoin(
                BASE_FACEBOOK_URL, "/%s/posts/%s" % (self.parent_handle, self.id)
            )

        if self.parent_id is not None:
            return urljoin(
                BASE_FACEBOOK_URL,
                "/permalink.php?story_fbid=%s&id=%s" % (self.id, self.parent_id),
            )

        if self.group_id is not None:
            return urljoin(
                BASE_FACEBOOK_URL, "/groups/%s/permalink/%s" % (self.group_id, self.id)
            )

        if self.group_handle is not None:
            return urljoin(
                BASE_FACEBOOK_URL,
                "/groups/%s/permalink/%s" % (self.group_handle, self.id),
            )

    @property
    def full_id(self):
        if self.parent_id is not None:
            return "%s_%s" % (self.parent_id, self.id)

        if self.group_id is not None:
            return "%s_%s" % (self.group_id, self.id)

        return None


class FacebookVideo(FacebookParsedItem):
    __slots__ = ("id", "parent_id")

    def __init__(self, video_id, parent_id=None):
        self.id = video_id
        self.parent_id = parent_id

    @property
    def url(self):
        if self.parent_id is None:
            return urljoin(BASE_FACEBOOK_URL, "/watch/?v=%s" % self.id)

        return urljoin(BASE_FACEBOOK_URL, "/%s/videos/%s" % (self.parent_id, self.id))


class FacebookPhoto(FacebookParsedItem):
    __slots__ = ("id", "group_id", "parent_id", "parent_handle", "album_id")

    def __init__(
        self, photo_id, group_id=None, parent_id=None, parent_handle=None, album_id=None
    ):
        self.id = photo_id
        self.group_id = group_id
        self.parent_id = parent_id
        self.parent_handle = parent_handle
        self.album_id = album_id

    @property
    def url(self):
        if self.group_id:
            return urljoin(
                BASE_FACEBOOK_URL,
                "/photo.php?fbid=%s&set=g.%s" % (self.id, self.group_id),
            )

        if self.parent_id:
            return urljoin(
                BASE_FACEBOOK_URL,
                "/%s/a.%s/%s" % (self.parent_id, self.album_id, self.id),
            )

        if self.parent_handle:
            return urljoin(
                BASE_FACEBOOK_URL,
                "/%s/a.%s/%s" % (self.parent_handle, self.album_id, self.id),
            )

        return urljoin(BASE_FACEBOOK_URL, "/photo.php?fbid=%s" % self.id)


def parse_facebook_url(url, allow_relative_urls=False):

    # Allowing relative urls scraped from facebook?
    if (
        allow_relative_urls
        and not url.startswith("http://")
        and not url.startswith("https://")
        and "facebook." not in url
    ):
        url = urljoin(BASE_FACEBOOK_URL, url)
    else:
        if not is_facebook_url(url):
            return None

    splitted = safe_urlsplit(url)

    if not splitted.path or splitted.path == "/":
        return None

    # Videos
    if "/watch" in splitted.path:
        query = safe_parse_qs(splitted.query)

        if "v" not in query:
            return None

        video_id = query["v"][0]

        return FacebookVideo(video_id)

    if "/videos/" in splitted.path:
        parts = pathsplit(splitted.path)

        return FacebookVideo(parts[2], parent_id=parts[0])

    # Photos
    if splitted.query and (
        splitted.path.endswith("/photo.php")
        or splitted.path.rstrip("/").endswith("/photo")
    ):
        query = safe_parse_qs(splitted.query)

        if "fbid" not in query:
            return None

        group_id = None
        album_id = None

        if "set" in query:
            sets = query["set"]

            group_id = next((s for s in sets if s.startswith("g.")), None)

            if group_id:
                group_id = group_id.split("g.", 1)[1]

            album_id = next((s for s in sets if s.startswith("a.")), None)

            if album_id:
                album_id = album_id.split("a.", 1)[1]

        return FacebookPhoto(query["fbid"][0], group_id=group_id, album_id=album_id)

    if "/photos/" in splitted.path:
        parts = pathsplit(splitted.path)

        parent_id_or_handle = parts[0]
        album_id = parts[2].replace("a.", "")
        photo_id = parts[3]

        if is_facebook_id(parent_id_or_handle):
            return FacebookPhoto(
                photo_id, album_id=album_id, parent_id=parent_id_or_handle
            )

        return FacebookPhoto(
            photo_id, album_id=album_id, parent_handle=parent_id_or_handle
        )

    # Obvious post path
    if "/posts/" in splitted.path:
        parts = pathsplit(splitted.path)

        if parts[0] == "groups":
            group_id_or_handle = parts[1]

            if NUMERIC_ID_RE.match(group_id_or_handle):
                return FacebookPost(parts[3], group_id=group_id_or_handle)
            return FacebookPost(parts[3], group_handle=group_id_or_handle)

        parent_id_or_handle = parts[0]

        if NUMERIC_ID_RE.match(parent_id_or_handle):
            return FacebookPost(parts[2], parent_id=parent_id_or_handle)

        return FacebookPost(parts[2], parent_handle=parent_id_or_handle)

    # Ye olded permalink path
    if splitted.query and (
        "/permalink.php" in splitted.path or "/story.php" in splitted.path
    ):
        query = safe_parse_qs(splitted.query)
        parent_id = query.get("id", None)

        if not parent_id:
            return None

        return FacebookPost(query["story_fbid"][0], parent_id=parent_id[0])

    # Group permalink path
    if "/groups/" in splitted.path:
        parts = pathsplit(splitted.path)

        if "/permalink/" in splitted.path:
            if is_facebook_id(parts[1]):
                return FacebookPost(parts[3], group_id=parts[1])

            return FacebookPost(parts[3], group_handle=parts[1])

        if is_facebook_id(parts[1]):
            return FacebookGroup(id=parts[1])

        return FacebookGroup(handle=parts[1])

    # Profile path
    if splitted.path == "/profile.php":
        query = safe_parse_qs(splitted.query)
        user_id = query["id"][0]
        return FacebookUser(user_id)

    # People path
    if splitted.path.startswith("/people"):
        parts = pathsplit(splitted.path)
        user_id = parts[2]
        return FacebookUser(user_id)

    # Handle path
    if splitted.path:
        parts = pathsplit(splitted.path)

        if not parts[0].endswith(".php"):
            return FacebookHandle(parts[0])

    return None


FACEBOOK_TYPES_HAVING_COMMENTS = (FacebookPost, FacebookPhoto, FacebookVideo)


def has_facebook_comments(url, allow_relative_urls=False):
    if not is_facebook_url(url):
        return False

    result = parse_facebook_url(url, allow_relative_urls=allow_relative_urls)

    return isinstance(result, FACEBOOK_TYPES_HAVING_COMMENTS)



# =============================================================================
# Ural Should Follow Href Function
# =============================================================================
#
# A function returning whether the given href attribute of a html <a> tag
# should be followed.
#
import re

HTTP_PROTOCOL_RE = re.compile(r'^https?://', re.I)

# NOTE: one should also compare the joined url to the current one usually
def should_follow_href(href):
    href = href.strip()

    if not href or href.startswith("#"):
        return False

    if ":" in href:
        return bool(HTTP_PROTOCOL_RE.match(href))

    return True



# -*- coding: utf-8 -*-
# =============================================================================
# Ural Could Be RSS URL
# =============================================================================
#
# Function returning whether the given url should be a rss feed url.
#
import re

from ural.utils import safe_urlsplit
from ural.get_hostname import get_hostname

HOSTNAME_RULES = [
    ("linkedin.com", "", False),
    ("news.google.com", "__i/rss", False),
    ("feeds.feedburner.com", "", True),
    ("feeds2.feedburner.com", "", True),
    ("news.google.com", "rss/topics", True),
]

QUERY_RE = re.compile(
    r"((page\=backend)|(?:feed|format|type)\=(?:xml|feed|atom|rss))", re.I
)
OBVIOUS_EXT_RE = re.compile(
    r"((^$)|(\/[^\.]*$)|(.+\.(?:xml|php|json|atom|rss)$))", re.I
)
FILE_RE = re.compile(
    r"(((?:feeds?|rss|atoms?|blogs?)\.(?:xml|atom|rss|php|json)$)|((?:news|latest|index|posts?)\.(?:xml|atom|rss)$)|(.+\.(?:rss|atom)$)|((?:[\/\.\-_]|^)(?:rss|feeds?|atom)[0-9]{,10}(?:[\/\.\-_]|$)))",
    re.I,
)


def could_be_rss(url):
    split = safe_urlsplit(url)
    hostname = get_hostname(url)

    if not hostname:
        return False

    for h, p, r in HOSTNAME_RULES:
        if hostname == h:
            if p in split.path:
                return r

            break

    if split.query and QUERY_RE.search(split.query):
        return True

    if split.path and not OBVIOUS_EXT_RE.match(split.path):
        return False

    if split.path and FILE_RE.search(split.path):
        return True

    return False



from ural.types import AnyUrlTarget

def fingerprint_url(url: str, unsplit: bool = ..., strip_suffix: bool = ...) -> str: ...
def fingerprint_hostname(hostname: str, strip_suffix: bool = ...) -> str: ...
def get_fingerprinted_hostname(
    url: AnyUrlTarget, infer_redirection: bool = True, strip_suffix: bool = ...
) -> str: ...



from typing import Optional
from ural.types import AnyUrlTarget

def is_facebook_id(value: str) -> bool: ...
def is_facebook_full_id(value: str) -> bool: ...
def is_facebook_url(url: AnyUrlTarget) -> bool: ...
def is_facebook_post_url(url: AnyUrlTarget) -> bool: ...
def is_facebook_link(url: AnyUrlTarget) -> bool: ...
def extract_url_from_facebook_link(url: str) -> Optional[str]: ...
def convert_facebook_url_to_mobile(url: str) -> str: ...

class FacebookParsedItem(object):
    @property
    def url(self) -> str: ...

class FacebookUser(FacebookParsedItem):
    id: str
    handle: Optional[str]

    def __init__(self, id: str, handle=None): ...

class FacebookHandle(FacebookParsedItem):
    handle: str

    def __init__(self, handle: str): ...

class FacebookGroup(FacebookParsedItem):
    id: str
    handle: Optional[str]

    def __init__(self, id: str, handle=None): ...

class FacebookPost(FacebookParsedItem):
    id: str
    parent_id: Optional[str]
    parent_handle: Optional[str]
    group_id: Optional[str]
    group_handle: Optional[str]

    def __init__(
        self,
        post_id: str,
        parent_id: Optional[str] = None,
        parent_handle: Optional[str] = None,
        group_id: Optional[str] = None,
        group_handle: Optional[str] = None,
    ): ...
    @property
    def full_id(self) -> Optional[str]: ...

class FacebookVideo(FacebookParsedItem):
    id: str
    video_id: Optional[str]
    parent_id: Optional[str]

    def __init__(self, video_id: str, parent_id: Optional[str] = None): ...

class FacebookPhoto(FacebookParsedItem):
    id: str
    group_id: Optional[str]
    parent_id: Optional[str]
    parent_handle: Optional[str]
    album_id: Optional[str]

    def __init__(
        self,
        photo_id: str,
        group_id: Optional[str] = None,
        parent_id: Optional[str] = None,
        parent_handle: Optional[str] = None,
        album_id: Optional[str] = None,
    ): ...

def parse_facebook_url(
    url: AnyUrlTarget, allow_relative_urls: bool = ...
) -> Optional[FacebookParsedItem]: ...
def has_facebook_comments(
    url: AnyUrlTarget, allow_relative_urls: bool = ...
) -> bool: ...



from typing import NamedTuple, Optional, Union
from ural.types import AnyUrlTarget

class InstagramUser(NamedTuple):
    name: str

class InstagramPost(NamedTuple):
    id: str
    name: str

class InstagramReel(NamedTuple):
    id: str

def is_instagram_post_shortcode(value: str) -> bool: ...
def is_instagram_username(value: str) -> bool: ...
def is_instagram_url(url: AnyUrlTarget) -> bool: ...
def parse_instagram_url(
    url: AnyUrlTarget,
) -> Optional[Union[InstagramUser, InstagramPost, InstagramReel]]: ...
def extract_username_from_instagram_url(url: AnyUrlTarget) -> Optional[str]: ...



from typing import Optional, List, Union, Dict, Callable, Any

from ural.types import QueryArgValue

FormatArgValueCallable = Callable[[str, QueryArgValue], Any]
Path = Union[str, List[Union[str, int]]]

def format_url(
    base_url: str,
    path: Optional[Path] = ...,
    args: Optional[Dict[str, Any]] = ...,
    format_arg_value: Optional[FormatArgValueCallable] = ...,
    fragment: Optional[str] = ...,
    ext: Optional[str] = ...,
) -> str: ...

class URLFormatter(object):
    BASE_URL = Optional[str]
    def __init__(
        self,
        base_url: Optional[str] = ...,
        path: Optional[Path] = ...,
        args: Optional[Dict[str, Any]] = ...,
        format_arg_value: Optional[FormatArgValueCallable] = ...,
        fragment: Optional[str] = ...,
        ext: Optional[str] = ...,
    ): ...
    def __call__(
        self,
        base_url: Optional[str] = ...,
        path: Optional[Path] = ...,
        args: Optional[Dict[str, Any]] = ...,
        format_arg_value: Optional[FormatArgValueCallable] = ...,
        fragment: Optional[str] = ...,
        ext: Optional[str] = ...,
    ) -> str: ...
    def format_arg_value(self, key: str, value: Any) -> QueryArgValue: ...
    def format(
        self,
        base_url: Optional[str] = ...,
        path: Optional[Path] = ...,
        args: Optional[Dict[str, Any]] = ...,
        format_arg_value: Optional[FormatArgValueCallable] = ...,
        fragment: Optional[str] = ...,
        ext: Optional[str] = ...,
    ) -> str: ...



from ural.utils import safe_urlsplit
from ural.has_special_host import is_special_host


class SuffixTrieNode(object):
    __slots__ = ("children", "exception", "leaf", "private")

    def __init__(self):
        self.children = None
        self.exception = None
        self.leaf = False
        self.private = False


class SuffixTrie(object):
    def __init__(self):
        self.__root = SuffixTrieNode()

    def add(self, suffix, private=False):
        node = self.__root

        # Iterating over the suffix parts in reverse order
        for part in reversed(suffix.split(".")):

            if part.startswith("!"):
                node.exception = part[1:]
                break

            # To save up some RAM, we initialize the children dict only
            # when strictly necessary
            if node.children is None:
                node.children = {}
                child = SuffixTrieNode()
                node.children[part] = child
            else:
                child = node.children.get(part)
                if child is None:
                    child = SuffixTrieNode()
                    node.children[part] = child

            node = child

        node.leaf = True

        if private:
            node.private = True

    def __walk(self, url):
        parsed = safe_urlsplit(url)

        if is_special_host(parsed.hostname):
            return None

        hostname = parsed.hostname.lower().rstrip(".")
        parts = hostname.split(".")

        current_length = 0
        suffix_length = 0
        match = None
        l = len(parts)

        node = self.__root

        for i in range(l - 1, -1, -1):
            part = parts[i]

            # Cannot go deeper
            if node.children is None:
                break

            # Exception
            if part == node.exception:
                break

            child = node.children.get(part)

            # Wildcards
            if child is None:
                child = node.children.get("*")

            # If the current part is not in current node's children, we can stop
            if child is None:
                break

            # Else we move deeper and increment our suffix offset
            current_length += 1
            node = child

            if node.leaf:
                suffix_length = current_length
                match = node

        # Checking the node we finished on is a leaf and is one we allow
        if match is None or not match.leaf:
            return None

        # hostname = suffix ?
        if l == suffix_length:
            offset = -1
        else:
            offset = max(1, l - suffix_length)

        return match, parts, offset

    def split(self, url):
        result = self.__walk(url)

        if result is None:
            return None

        _, parts, offset = result

        # hostname = suffix
        if offset < 0:
            return "", ".".join(parts)

        return ".".join(parts[:offset]), ".".join(parts[offset:])

    def has_valid_domain_name(self, url):
        result = self.__walk(url)

        if result is None:
            return False

        return True

    def extract_suffix(self, url):
        result = self.__walk(url)

        if result is None:
            return None

        # TODO: we can restrict to public or private here easily
        _, parts, offset = result

        # hostname = suffix
        if offset < 0:
            return ".".join(parts)

        return ".".join(parts[offset:])

    def extract_domain_name(self, url):
        result = self.__walk(url)

        if result is None:
            return None

        # TODO: we can restrict to public or private here easily
        _, parts, offset = result

        # hostname = suffix
        if offset < 0:
            return ".".join(parts)

        return ".".join(parts[offset - 1 :])



from typing import List, Tuple, Optional

def safely_unquote_auth_item(string: str) -> str: ...
def safely_unquote_path(string: str) -> str: ...
def safely_unquote_query_item(string: str) -> str: ...
def safely_unquote_fragment(string: str) -> str: ...
def safely_unquote_qsl(
    qsl: List[Tuple[str, Optional[str]]]
) -> List[Tuple[str, Optional[str]]]: ...
def safely_quote(string: str) -> str: ...
def safely_quote_qsl(
    qsl: List[Tuple[str, Optional[str]]]
) -> List[Tuple[str, Optional[str]]]: ...
def upper_quoted(string: str) -> str: ...



from ural.utils import (
    urlsplit,
    urlunsplit,
    SplitResult,
    unsplit_netloc,
    safe_qsl_iter,
    safe_serialize_qsl,
    decode_punycode_hostname,
    normpath,
)
from ural.quote import (
    safely_unquote_auth_item,
    safely_unquote_path,
    safely_unquote_qsl,
    safely_unquote_fragment,
    safely_quote,
    safely_quote_qsl,
    upper_quoted,
)
from ural.ensure_protocol import ensure_protocol
from ural.patterns import CONTROL_CHARS_RE


def canonicalize_url(
    url, default_protocol="https", unsplit=True, quoted=False, strip_fragment=False
):
    # Cleaning
    url = CONTROL_CHARS_RE.sub("", url)
    url = url.strip()
    url = upper_quoted(url)

    # Ensuring a protocol
    url = ensure_protocol(url, default_protocol)

    # Parsing
    splitted = urlsplit(url)
    scheme, netloc, path, query, fragment = splitted
    user, password, hostname, port = (
        splitted.username,
        splitted.password,
        splitted.hostname,
        splitted.port,
    )

    # Decoding and normalizing hostname
    if hostname:
        hostname = decode_punycode_hostname(hostname)
        hostname = hostname.lower()

    # Dropping HTTP/HTTPS ports
    if port == 80 or port == 443:
        port = None

    if strip_fragment:
        fragment = None

    # Empty path etc.
    if not path or path == "/":
        if not query and not fragment:
            path = ""
        else:
            path = "/"

    # Path normalization
    else:
        path = normpath(path)

    # Quotes
    if user:
        if quoted:
            user = safely_quote(user)
        else:
            user = safely_unquote_auth_item(user)

    if password:
        if quoted:
            password = safely_quote(password)
        else:
            password = safely_unquote_auth_item(password)

    if quoted:
        path = safely_quote(path)
    else:
        path = safely_unquote_path(path)

    qsl = safe_qsl_iter(query)

    if quoted:
        qsl = safely_quote_qsl(qsl)
    else:
        qsl = safely_unquote_qsl(qsl)

    query = safe_serialize_qsl(qsl)

    if fragment:
        if quoted:
            fragment = safely_quote(fragment)
        else:
            fragment = safely_unquote_fragment(fragment)

    # Repacking
    netloc = unsplit_netloc(user, password, hostname, port)
    result = SplitResult(scheme, netloc, path, query, fragment)

    if not unsplit:
        return result

    # Serializing
    return urlunsplit(result)



# =============================================================================
# Ural Hostname Trie Set
# =============================================================================
#
# Hostname-aware custom implementation of a prefix trie used to efficiently
# match large numbers of domains.
#
from ural.classes import TrieDict
from ural.utils import safe_urlsplit, decode_punycode_hostname
from ural.has_special_host import is_special_host


def tokenize_hostname(hostname):
    if is_special_host(hostname):
        return [hostname]

    hostname = hostname.strip().lower()
    hostname_parts = decode_punycode_hostname(hostname, as_parts=True)

    return reversed(hostname_parts)


def join_hostname(prefix):
    return ".".join(reversed(prefix))


# NOTE: this trie currently has undefined behavior with some special hosts
class HostnameTrieSet(object):
    def __init__(self):
        self.__trie = TrieDict()

    def __len__(self):
        return len(self.__trie)

    def add(self, hostname):

        prefix = list(tokenize_hostname(hostname))

        self.__trie.set_and_prune_if_shorter(prefix, True)

    def match(self, url):
        url = safe_urlsplit(url)

        if not url.hostname:
            return False

        prefix = tokenize_hostname(url.hostname)

        return bool(self.__trie.longest_matching_prefix_value(prefix))

    def __iter__(self):
        for prefix in self.__trie.prefixes():
            yield join_hostname(prefix)



import sys
from typing import Union, Optional
from urllib.parse import SplitResult

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

AnyUrlTarget = Union[str, SplitResult]
QueryArgValue = Optional[Union[str, int, float, bool]]



from typing import NamedTuple, Union, Optional
from ural.types import AnyUrlTarget

class TwitterTweet(NamedTuple):
    user_screen_name: str
    id: str

class TwitterUser(NamedTuple):
    screen_name: str

class TwitterList(NamedTuple):
    id: str

def is_twitter_url(url: AnyUrlTarget) -> bool: ...
def normalize_screen_name(username: str) -> str: ...
def parse_twitter_url(
    url: AnyUrlTarget,
) -> Optional[Union[TwitterTweet, TwitterUser, TwitterList]]: ...
def extract_screen_name_from_twitter_url(url: AnyUrlTarget) -> Optional[str]: ...



from typing import Optional, Dict, List, Tuple, Iterator, Iterable, Union, overload
from urllib.parse import SplitResult
from ural.types import AnyUrlTarget, Literal, QueryArgValue

def urlsplit(url: str) -> SplitResult: ...
def safe_urlsplit(url: AnyUrlTarget, scheme: Optional[str] = ...) -> SplitResult: ...
def pathsplit(urlpath: str) -> List[str]: ...
def urlpathsplit(url: AnyUrlTarget) -> List[str]: ...
def normpath(urlpath: str, drop_consecutive_slashes: bool = True) -> str: ...
def attempt_to_decode_idna(string: str) -> str: ...
@overload
def decode_punycode_hostname(hostname: str, as_parts: Literal[False] = ...) -> str: ...
@overload
def decode_punycode_hostname(
    hostname: str, as_parts: Literal[True] = ...
) -> List[str]: ...
def fix_common_query_mistakes(query: str) -> str: ...
def safe_parse_qs(query: str) -> Dict[str, List[str]]: ...
def add_query_argument(
    url: str, name: str, value: QueryArgValue, quote: bool = ...
) -> str: ...
def unsplit_netloc(
    username: Optional[str],
    password: Optional[str],
    hostname: Optional[str],
    port: Optional[Union[str, int]],
) -> str: ...
def safe_qsl_iter(query: str) -> Iterator[Tuple[str, Optional[str]]]: ...
def safe_serialize_qsl(qsl: Iterable[Tuple[str, Optional[str]]]): ...



from typing import Union, Optional, Callable, overload
from ural.types import AnyUrlTarget, Literal
from urllib.parse import SplitResult

@overload
def normalize_url(
    url: str,
    sort_query: bool = ...,
    strip_authentication: bool = ...,
    strip_trailing_slash: bool = ...,
    strip_index: bool = ...,
    strip_protocol: bool = ...,
    strip_irrelevant_subdomains: bool = ...,
    strip_fragment: Union[bool, Literal["except-routing"]] = ...,
    normalize_amp: bool = ...,
    fix_common_mistakes: bool = ...,
    infer_redirection: bool = ...,
    quoted: bool = ...,
    #
    unsplit: Literal[True] = ...,
    query_item_filter: Optional[Callable[[str, str], bool]] = ...,
) -> str: ...
@overload
def normalize_url(
    url: str,
    sort_query: bool = ...,
    strip_authentication: bool = ...,
    strip_trailing_slash: bool = ...,
    strip_index: bool = ...,
    strip_protocol: bool = ...,
    strip_irrelevant_subdomains: bool = ...,
    strip_fragment: Union[bool, Literal["except-routing"]] = ...,
    normalize_amp: bool = ...,
    fix_common_mistakes: bool = ...,
    infer_redirection: bool = ...,
    quoted: bool = ...,
    #
    unsplit: Literal[False] = ...,
    query_item_filter: Optional[Callable[[str, str], bool]] = ...,
) -> SplitResult: ...
def normalize_hostname(hostname: str, normalize_amp: bool = True) -> str: ...
def get_normalized_hostname(
    url: AnyUrlTarget,
    normalize_amp: bool = True,
    infer_redirection: bool = True,
) -> str: ...



# =============================================================================
# Ural Could Be HTML Function
# =============================================================================
#
# A function returning whether the given url could point to a HTML document.
#
from os.path import splitext
from ural.utils import safe_urlsplit

HTML_LIKE_EXTENSIONS = {
    ".htm",
    ".html",
    ".dhtml",
    ".shtml",
    ".xhtml",
    ".asp",
    ".aspx",
    ".jsp",
    ".jspx",
    ".pl",
    ".php",
    ".php5",
    ".cgi",
    ".bin",
}


def could_be_html(url):
    path = safe_urlsplit(url).path

    _, ext = splitext(path)

    if not ext:
        return True

    return ext in HTML_LIKE_EXTENSIONS



from ural.utils import string_type, urlunsplit
from ural.lru.serialization import serialize_lru, unserialize_lru
from ural.lru.stems import lru_stems


def url_to_lru(url, suffix_aware=False):
    return serialize_lru(lru_stems(url, suffix_aware=suffix_aware))


def lru_to_url(lru):

    # Handling both stems and serialized lru
    if isinstance(lru, string_type):
        stems = unserialize_lru(lru)
    else:
        stems = lru

    # Indexing stems
    stems_index = {}

    for stem in stems:
        tag, value = stem.split(":", 1)

        if tag == "h" and tag in stems_index:
            stems_index[tag] = value + "." + stems_index[tag]
        elif tag == "p" and tag in stems_index:
            stems_index[tag] += "/" + value
        else:
            stems_index[tag] = value

    # Building the url back
    scheme = stems_index.get("s", "")
    auth = stems_index.get("u", "")

    w = stems_index.get("w")

    if w is not None:
        auth += ":" + w

    netloc = ""

    if auth:
        netloc = auth + "@"

    netloc += stems_index.get("h", "")

    t = stems_index.get("t")

    if t is not None:
        netloc += ":" + t

    p = stems_index.get("p")

    path = ""

    if p is not None:
        path = "/" + p

    query = stems_index.get("q", "")
    fragment = stems_index.get("f", "")

    return urlunsplit((scheme, netloc, path, query, fragment))



from typing import Optional, Tuple
from ural.types import AnyUrlTarget

def upgrade(transient: bool = ...) -> None: ...
def get_domain_name(url: AnyUrlTarget) -> Optional[str]: ...
def has_valid_suffix(url: AnyUrlTarget) -> bool: ...
def split_suffix(url: AnyUrlTarget) -> Optional[Tuple[str, str]]: ...
def is_valid_tld(tld: str) -> bool: ...
def has_valid_tld(url: AnyUrlTarget) -> bool: ...



from typing import Tuple, Optional
from ural.types import AnyUrlTarget

class SuffixTrie(object):
    def add(self, tld: str, private: bool = ...) -> None: ...
    def split(self, url: AnyUrlTarget) -> Optional[Tuple[str, str]]: ...
    def has_valid_domain_name(self, url: AnyUrlTarget) -> bool: ...
    def extract_suffix(self, url: AnyUrlTarget) -> Optional[str]: ...
    def extract_domain_name(self, url: AnyUrlTarget) -> Optional[str]: ...



from ural.data import ISO_3166_1_COUNTRIES_ALPHA_2
from ural.normalize_url import normalize_url, normalize_hostname
from ural.utils import SplitResult, urlunsplit, urlsplit, unsplit_netloc
from ural.infer_redirection import infer_redirection as resolve
from ural.ensure_protocol import ensure_protocol
from ural.tld import split_suffix

LANG_QUERY_KEYS = ("gl", "hl")

# TODO: drop tld


def lang_query_item_filter(key, _):
    return key not in LANG_QUERY_KEYS


def strip_lang_subdomains_from_hostname(hostname):
    if hostname.count(".") > 1:
        subdomain, remaining_hostname = hostname.split(".", 1)
        if len(subdomain) == 5 and "-" in subdomain:
            lang, country = subdomain.split("-", 1)
            if len(lang) == 2 and len(country) == 2:
                if (
                    lang.upper() in ISO_3166_1_COUNTRIES_ALPHA_2
                    and country.upper() in ISO_3166_1_COUNTRIES_ALPHA_2
                ):
                    hostname = remaining_hostname
        elif len(subdomain) == 2:
            if subdomain.upper() in ISO_3166_1_COUNTRIES_ALPHA_2:
                hostname = remaining_hostname

    return hostname


def fingerprint_hostname(hostname, strip_suffix=False):
    hostname = normalize_hostname(hostname)

    if strip_suffix:
        # TODO: this is not performant because the code path reparses again
        r = split_suffix(hostname)

        if r is not None:
            hostname, _ = r

    return strip_lang_subdomains_from_hostname(hostname)


def get_fingerprinted_hostname(url, infer_redirection=True, strip_suffix=False):
    if infer_redirection:
        url = resolve(url)

    if isinstance(url, SplitResult):
        splitted = url
    else:
        try:
            splitted = urlsplit(ensure_protocol(url.strip()))
        except ValueError:
            return None

    if not splitted.hostname:
        return None

    return fingerprint_hostname(splitted.hostname, strip_suffix=strip_suffix)


def fingerprint_url(url, unsplit=True, strip_suffix=False):
    url = url.lower()

    splitted = normalize_url(
        url, unsplit=False, query_item_filter=lang_query_item_filter
    )
    _, netloc, path, query, fragment = splitted

    user, password, hostname, port = (
        splitted.username,
        splitted.password,
        splitted.hostname,
        splitted.port,
    )

    if hostname:
        hostname = strip_lang_subdomains_from_hostname(hostname)

        if strip_suffix:
            # TODO: this is not performant because the code path reparses again
            r = split_suffix(hostname)

            if r is not None:
                hostname, _ = r

    # Dropping port
    port = None

    netloc = unsplit_netloc(user, password, hostname, port)

    result = SplitResult("", netloc, path, query, fragment)

    if not unsplit:
        return result

    return urlunsplit(result)[2:]



# =============================================================================
# Ural Google-related heuristic functions
# =============================================================================
#
# Collection of functions related to Google urls.
#
import re
from ural.utils import safe_urlsplit, unquote, pathsplit
from ural.patterns import QUERY_VALUE_IN_URL_TEMPLATE

AMP_QUERY_RE = re.compile(r"amp(_.+)=?", re.I)
AMP_SUFFIXES_RE = re.compile(r"(?:\.amp(?=\.html$)|\.amp/?$|(?<=/)amp/?$)", re.I)

URL_EXTRACT_RE = re.compile(QUERY_VALUE_IN_URL_TEMPLATE % r"url")

DRIVE_TYPES = ["document", "presentation", "spreadsheets"]


def is_amp_url(url):
    splitted = safe_urlsplit(url)

    if splitted.hostname.endswith(".ampproject.org"):
        return True

    if splitted.hostname.startswith("amp-"):
        return True

    if splitted.hostname.startswith("amp."):
        return True

    if "/amp/" in splitted.path:
        return True

    if AMP_SUFFIXES_RE.search(splitted.path):
        return True

    if splitted.query and AMP_QUERY_RE.search(splitted.query):
        return True

    return False


def is_google_link(url):
    splitted = safe_urlsplit(url)

    if not splitted.hostname or "google." not in splitted.hostname:
        return False

    if splitted.path != "/url":
        return False

    return True


def extract_url_from_google_link(url):
    m = URL_EXTRACT_RE.search(url)

    if m is None:
        return None

    return unquote(m.group(1))


class GoogleDriveParsedItem(object):
    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False

        for attr in self.__slots__:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __repr__(self):
        class_name = self.__class__.__name__

        representation = "<" + class_name

        for key in self.__slots__:
            value = getattr(self, key)

            if value is None:
                continue

            representation += " %s=%s" % (key, value)

        representation += ">"

        return representation


class GoogleDriveFile(GoogleDriveParsedItem):
    __slots__ = ("type", "id")

    def __init__(self, _type, _id):
        self.type = _type
        self.id = _id

    @property
    def url(self):
        return "https://docs.google.com/%s/d/%s" % (self.type, self.id)

    def get_export_url(self, format="csv"):
        return "https://docs.google.com/%s/d/%s/export?exportFormat=%s" % (
            self.type,
            self.id,
            format,
        )


class GoogleDrivePublicLink(GoogleDriveParsedItem):
    __slots__ = ("type", "id")

    def __init__(self, _type, _id):
        self.type = _type
        self.id = _id

    @property
    def url(self):
        return "https://docs.google.com/%s/d/e/%s/pub" % (self.type, self.id)

    def get_export_url(self, format="csv"):
        return "https://docs.google.com/%s/d/e/%s/pub?output=%s" % (
            self.type,
            self.id,
            format,
        )


def parse_google_drive_url(url):
    splitted = safe_urlsplit(url)

    if "docs.google.com" not in splitted.netloc:
        return None

    path = pathsplit(splitted.path)

    if len(path) < 3:
        return None

    drive_type = path[0]

    if drive_type not in DRIVE_TYPES:
        return None

    if path[1] != "d":
        return None

    if path[-1] == "pub":
        if path[2] != "e":
            return None

        return GoogleDrivePublicLink(drive_type, path[3])

    return GoogleDriveFile(drive_type, path[2])


def extract_id_from_google_drive_url(url):
    parsed = parse_google_drive_url(url)

    if isinstance(parsed, GoogleDriveFile):
        return parsed.id

    return None



from typing import Optional
from ural.types import AnyUrlTarget

def is_amp_url(url: AnyUrlTarget) -> bool: ...
def is_google_link(url: AnyUrlTarget) -> bool: ...
def extract_url_from_google_link(url: str) -> Optional[str]: ...

class GoogleDriveParsedItem(object):
    type: str
    id: str

    def __init__(self, _type: str, _id: str): ...
    @property
    def url(self) -> str: ...
    def get_export_url(self, format: str = ...) -> str: ...

class GoogleDriveFile(GoogleDriveParsedItem): ...
class GoogleDrivePublicLink(GoogleDriveParsedItem): ...

def parse_google_drive_url(url: AnyUrlTarget) -> Optional[GoogleDriveParsedItem]: ...
def extract_id_from_google_drive_url(url: AnyUrlTarget) -> Optional[str]: ...



from typing import NamedTuple, Optional, Union
from ural.types import AnyUrlTarget

class YoutubeVideo(NamedTuple):
    id: str

class YoutubeUser(NamedTuple):
    id: str
    name: str

class YoutubeChannel(NamedTuple):
    id: str
    name: str

def is_youtube_url(url: AnyUrlTarget) -> bool: ...
def is_youtube_video_id(value: str) -> bool: ...
def parse_youtube_url(
    url: AnyUrlTarget, fix_common_mistakes: bool = ...
) -> Optional[Union[YoutubeVideo, YoutubeUser, YoutubeChannel]]: ...
def extract_video_id_from_youtube_url(url: AnyUrlTarget) -> Optional[str]: ...
def normalize_youtube_url(url: AnyUrlTarget) -> str: ...



# =============================================================================
# Ural LRU Module
# =============================================================================
#
# Module gathering LRU-related functions.
#
from ural.lru.conversion import url_to_lru, lru_to_url
from ural.lru.serialization import serialize_lru, unserialize_lru
from ural.lru.stems import lru_stems, normalized_lru_stems
from ural.lru.trie import LRUTrie, NormalizedLRUTrie

__all__ = [
    "serialize_lru",
    "unserialize_lru",
    "lru_stems",
    "normalized_lru_stems",
    "LRUTrie",
    "NormalizedLRUTrie",
    "url_to_lru",
    "lru_to_url",
]
